//
// rgsproc.cc
//
// Authors:  Marco Beijersbergen (cosine), Giuseppe Vacanti (cosine),
//           Peter Videler (CARA), John Doug Reynolds (CAL)
//           ESA XMM SOC
//         
// Description: Pipeline processing metatask for RGS
//
// $Id: rgsproc.cc,v 1.79 2000/08/14 15:55:21 pvideler Exp $

#include <set>		// C++
#include <algorithm>
#include <strstream>
#include <unistd.h>

#include <DalVec.h>	// SAS
#include <OalServer.h>
#include <metatask.h>
#include <errstr.h>
#include <param.h>

#include <atthkgen.h>	// tasks
#include <rgsoffsetcalc.h>
#include <rgssources.h>
#include <rgsframes.h>
#include <rgsbadpix.h>
#include <rgsevents.h>
#include <evlistcomb.h>
#include <gtimerge.h>
#include <rgsangles.h>
#include <rgsfilter.h>
#include <rgsregions.h>
#include <rgsspectrum.h>

#include "rgsproc.h"

// flow control
static bool processevents, processsources, processexposure, processregions;
static vector<int> orders;

// rgssources
static bool withepicset,withsrc;
static string epicset,attstyle,srclabel,srcstyle;
static double attra,attdec,attapos,srcrate,srcdisp,srcxdsp,srcra,srcdec;

// atthkgen
static bool withattitude;
static string attitude;
static double timestep;

// rgsoffsetcalc
static bool calcoffsets, withoffsethistogram;

// rgsbadpix
static bool withadvisory, withfoundhot;
static short pixnoiselimit, colnoiselimit;
static float pixsharpness, colsharpness;

// rgsevents
static string detcoord;
static bool reconstruct, withdetcoord, withcti, withgain;

// rgsangles
static string betabinning,xdispbinning;
static double betabinref,xdispbinref,betabinwidth,xdispbinwidth,betamin,betamax,xdispmin,xdispmax,betaminwarn,betamaxwarn,xdispminwarn,xdispmaxwarn;
static int nbetabins,nxdispbins;
static bool usebinning,withpointingcolumn,withmlambdacolumn;

// gtimerge
static bool withauxgti;
static string gtitable;

// rgsfilter
static vector<string> rejflags;
static bool withcombmap;

// rgsregions
static int spatialres, orderres;
static double xpsfincl, xpsfexcl, pdistincl;
static string procsrcsexpr, exclsrcsexpr;

// rgsspectrum
static int rebin, edgechannels;

static vector<string> stringVector( char *strs[], size_t size )
{
   vector<string> v(size);
   for ( unsigned i = 0; i < size; ++i ) v[i] = strs[i];
   return v;
}
#define STRINGVECTOR(array) stringVector(array,sizeof(array)/sizeof(*array))

static string basename(const OdfSet *odfSet,
	Xmm::Instrument instr=Xmm::INSTRUMENT_NONE)
{
  PROC
  string base = odfSet->basename().substr(0,24);
  string instrument = "";
  switch (instr) {
    case Xmm::EMOS1: instrument = "M1"; break;
    case Xmm::EMOS2: instrument = "M2"; break;
    case Xmm::EPN: instrument = "PN"; break;
    case Xmm::RGS1: instrument = "R1"; break;
    case Xmm::RGS2: instrument = "R2"; break;
    case Xmm::OM: instrument = "OM"; break;
    case Xmm::SC: instrument = "SC"; break;
    case Xmm::INSTRUMENT_NONE:
    case Xmm::XRT1:
    case Xmm::XRT2:
    case Xmm::XRT3:
    case Xmm::ERM: ;
  }
  if (instrument != "")
    base = base.substr(0,16) + instrument + "X00000";
  RETURN base;
}

static string process( const OdfSet* odfSet, const string &offsetset )
{
   PROC

   string eventset = basename(odfSet) + "_events.ds";

   openParameters("rgsframes");
   setParameter("set", odfSet->name());
   setParameter("outset", eventset);
   task(rgsframes);
   closeParameters();

   openParameters("rgsbadpix");
   setParameter("set", eventset);
   setParameter("withadvisory",withadvisory);
   setParameter("withfoundhot",withfoundhot);
   setParameter("pixnoiselimit",pixnoiselimit);
   setParameter("colnoiselimit",colnoiselimit);
   setParameter("pixsharpness",pixsharpness);
   setParameter("colsharpness",colsharpness);
   task(rgsbadpix);
   closeParameters();

   openParameters("rgsevents");
   setParameter("set", eventset);
   setParameter("offset", offsetset);
   setParameter("withdetcoord",withdetcoord);
   setParameter("withoffset", calcoffsets);
   setParameter("withcti",withcti);
   setParameter("withgain",withgain);
   setParameter("detcoord", detcoord);
   setParameter("reconstruct", reconstruct);
   task(rgsevents);
   closeParameters();

   RETURN eventset;
}

static void process( const Exposure* exposure, const string &offsetset )
{
  PROC

  // Copy the list of ODF files in this exposures.
  typedef list<const OdfSet*> OFList;
  OFList odfSets = exposure->sets();

  if (odfSets.size() == 0) ALTRETURN;

  // We need the first CCD file to know the X or U status and the ODF exposure
  // number.
  const OdfSet *firstSet = *odfSets.begin();
  string base = basename(firstSet).substr(0,22);
  string mergedset   = base + "00_merged.ds";
  string filteredset = base + "00_filtered.ds";
  string sourcesset  = base + "00_sources.ds";
  string sourcestable = sourcesset + ":SRCLIST";

  if ( processevents ) {
     string instrument;
     switch ( oalServer.instrument() ) {
     case Xmm::RGS1: instrument = "rgs1"; break;
     case Xmm::RGS2: instrument = "rgs2"; break;
     default: errstr << "invalid instrument " << oalServer.instrument()
		     << Sas::error("invalidInstrument");
     }

     openParameters("rgssources");
     setParameter("filemode","create");
     setParameter("primestyle","auto");
     setParameter("changeattitude",true);
     setParameter("changeprime",true);
     setParameter("instrument",instrument);
     setParameter("exposure",(int)firstSet->exposureNr());
     setParameter("rgsset",sourcesset);
     setParameter("withepicset",withepicset);
     if ( withepicset ) {
	setParameter("epicset",epicset);
	setParameter("enablefilter",true);
     }

     setParameter("atthkset",attitude);
     setParameter("attstyle",attstyle);
     setParameter("attra",attra);
     setParameter("attdec",attdec);
     setParameter("attapos",attapos);

     setParameter("addusersource",withsrc);
     setParameter("userasprime",true);
     setParameter("label",srclabel);
     setParameter("rate",srcrate);
     setParameter("positionstyle",srcstyle);
     setParameter("deltadisp",srcdisp);
     setParameter("deltaxdsp",srcxdsp);
     setParameter("ra",srcra);
     setParameter("dec",srcdec);
     task(rgssources);
     closeParameters();

     // Loop over the event files.
     vector<string> evlist;

     OFList::const_iterator ifn;
     for ( ifn = odfSets.begin(); ifn != odfSets.end(); ++ifn ) {
	const OdfSet* odfSet = (*ifn);

	errstr << "considering " << odfSet->name()
	       << Sas::message(Msg::Meta,Msg::Verbose);

	if ( odfSet->type() == odfSet->EVENT_LIST )
	   evlist.push_back( process(odfSet,offsetset) );
     }
  
     char *mainattributes[] = {
	"CCDOCB", "SOURCEID", "DATE-OBS", "DATE-END", "DATATYPE",
	"TSTART", "TSTOP", "TELAPS", "XMMEA_RG"
     };
     char *othertables[] = {
	"STDGTI", "BADPIX0", "BADPIX1", "REJPIX0", "REJPIX1", "EXPOSURE"
     };
     char *rgsimgcolnames[] = {
	"TIME", "FLAG", "BETA", "XDSP", "CHIPX", "CHIPY",
	"PHA", "SHAPE", "GRADE", "PI"
     };
     char *rgsimgcoltypes[] = {
	"double", "int32", "single", "single", "int16", "int16",
	"int16", "int8", "int8", "int16"
     };

     openParameters("evlistcomb");
     setParameter("eventsets",evlist);
     setParameter("imagingset",mergedset);
     setParameter("instrument","rgs");
     setParameter("mainattributes",STRINGVECTOR(mainattributes));
     setParameter("othertables",STRINGVECTOR(othertables));
     setParameter("rgsimgcolnames",STRINGVECTOR(rgsimgcolnames));
     setParameter("rgsimgcoltypes",STRINGVECTOR(rgsimgcoltypes));
     task(evlistcomb);
     closeParameters();

     if ( dataSetServer->exists(mergedset) ) {
	struct DataSet *dataset;
	vector<string> names, tables;

	dataset = dataSetServer->open(mergedset,Dal::Read);
	for ( unsigned i = 1; i <= 9; ++i ) {
	   string extname = string("STDGTI0") + char('0'+i);
	   if ( dataset->hasBlock(extname) ) {
	      names.push_back(extname);
	      tables.push_back(mergedset+":"+extname);
	   }
	}
	dataSetServer->close(dataset);

	openParameters("gtimerge");
	setParameter("gtitable",mergedset+":STDGTI");
	setParameter("tables",tables);
	setParameter("mergemode","and");
	task(gtimerge);
	closeParameters();

	dataset = dataSetServer->open(mergedset,Dal::Modify);
	vector<string>::const_iterator itr;
	for ( itr = names.begin(); itr != names.end(); ++itr )
	   dataset->deleteBlock(*itr);
	dataSetServer->close(dataset);
     }
     else {
	errstr << "no events were found" << Sas::warning("noEvents");
	return;
     }
  }

  if ( processsources ) {
     openParameters("rgsangles");
     setParameter("sourceset",sourcestable);
     setParameter("set", mergedset);
     setParameter("betabinning",betabinning);
     setParameter("betabinref",betabinref);
     setParameter("betabinwidth",betabinwidth);
     setParameter("nbetabins",nbetabins);
     setParameter("betamin",betamin);
     setParameter("betamax",betamax);
     setParameter("xdispbinning",xdispbinning);
     setParameter("xdispbinref",xdispbinref);
     setParameter("xdispbinwidth",xdispbinwidth);
     setParameter("nxdispbins",nxdispbins);
     setParameter("xdispmin",xdispmin);
     setParameter("xdispmax",xdispmax);
     setParameter("withpointingcolumn",withpointingcolumn);
     setParameter("withmlambdacolumn",withmlambdacolumn);
     task(rgsangles);
     closeParameters();
  }

  if ( processexposure ) {
     openParameters("rgsfilter");
     setParameter("set",mergedset);
     setParameter("filteredset",filteredset);
     setParameter("rejflags",rejflags);
     setParameter("withauxgti",withauxgti);
     setParameter("gtitable",gtitable);
     setParameter("withexpmaps",true);
     setParameter("withcombmap",withcombmap);
     setParameter("expimageset",base+"00_expmap.ds");
     task(rgsfilter);
     closeParameters();
  }

  if (!processregions) return;

  // Construct all required extraction regions

  int maxOrder = 1;
  vector<int>::const_iterator iter;
  for ( iter = orders.begin(); iter != orders.end(); ++iter )
     if ( *iter > maxOrder ) maxOrder = *iter;

  openParameters("rgsregions");
  setParameter("spatialres",spatialres);
  setParameter("orderres",orderres);
  setParameter("xpsfbelow",xpsfincl);
  setParameter("xpsfabove",xpsfincl);
  setParameter("xpsfexcl",xpsfexcl);
  setParameter("pdistincl",pdistincl);
  setParameter("procsrcsexpr",procsrcsexpr);
  setParameter("exclsrcsexpr",exclsrcsexpr);
  setParameter("setflags",true);
  setParameter("maxorder",maxOrder);
  setParameter("eventset",filteredset);
  setParameter("set",sourcesset);
  task(rgsregions);
  closeParameters();

  // Produce background spectrum for primary source

  DataSet* set = dataSetServer->open(sourcesset,Dal::Read);
  const Table* table = set->table(sourcestable);
  int primesrc = table->attribute("PRIMESRC")->asInt();

  for ( iter = orders.begin(); iter != orders.end(); ++iter ) {
     strstream ss;
     ss << base << "00_bkg" << primesrc << "o" << *iter << ".pha" << ends;

     openParameters("rgsspectrum");
     setParameter("rebin",rebin);
     setParameter("edgechannels",edgechannels);
     setParameter("sourceset",sourcesset);
     setParameter("source",primesrc);
     setParameter("order",*iter);
     setParameter("set",filteredset);
     setParameter("file",ss.str());
     setParameter("type","bkg");
     task(rgsspectrum);
     closeParameters();
  }

  // Produce selected foreground spectra

  {
     const DalVecL process( table->column("PROCESS") );
     const DalVecI index( table->column("INDEX") );

     for ( unsigned row = 0; row < table->rows(); ++row )
	if ( process[row] )
	   for ( iter = orders.begin(); iter != orders.end(); ++iter ) {
	      strstream ss;
	      ss << base << "00_net" << index[row] << "o" << *iter << ".pha"
		 << ends;

	      openParameters("rgsspectrum");
	      setParameter("rebin",rebin);
	      setParameter("edgechannels",edgechannels);
	      setParameter("sourceset",sourcesset);
	      setParameter("source",index[row]);
	      setParameter("order",*iter);
	      setParameter("set",filteredset);
	      setParameter("file",ss.str());
	      setParameter("type","net");
	      task(rgsspectrum);
	      closeParameters();
	   }
  }

  dataSetServer->close(set);
  ENDPROC
}

void rgsproc ()
{
  PROC

  // flow control
  processevents = booleanParameter("processevents");
  processsources = booleanParameter("processsources");
  processexposure = booleanParameter("processexposure");
  processregions = booleanParameter("processregions");
  orders = intListParameter("orders");

  // rgssources
  withepicset = booleanParameter("withepicset");
  withsrc = booleanParameter("withsrc");
  epicset = stringParameter("epicset");
  attstyle = stringParameter("attstyle");
  srclabel = stringParameter("srclabel");
  srcstyle = stringParameter("srcstyle");
  attra = realParameter("attra");
  attdec = realParameter("attdec");
  attapos = realParameter("attapos");
  srcrate = realParameter("srcrate");
  srcdisp = realParameter("srcdisp");
  srcxdsp = realParameter("srcxdsp");
  srcra = realParameter("srcra");
  srcdec = realParameter("srcdec");

  // atthkgen
  withattitude = booleanParameter("withattitude");
  attitude = stringParameter("attitude");
  timestep = realParameter("timestep");

  // rgsoffsetcalc
  calcoffsets = booleanParameter("calcoffsets");
  withoffsethistogram = booleanParameter("withoffsethistogram");

  // rgsbadpix
  withadvisory = booleanParameter("withadvisory");
  withfoundhot = booleanParameter("withfoundhot");
  pixnoiselimit = intParameter("pixnoiselimit");
  colnoiselimit = intParameter("colnoiselimit");
  pixsharpness = realParameter("pixsharpness");
  colsharpness = realParameter("colsharpness");

  // rgsevents
  withdetcoord = booleanParameter("withdetcoord");
  detcoord = stringParameter("detcoord");
  reconstruct = booleanParameter("reconstruct");
  withgain = booleanParameter("withgain");
  withcti = booleanParameter("withcti");

  // rgsangles
  usebinning = booleanParameter("usebinning");
  betabinning = stringParameter("betabinning");
  betabinwidth = realParameter("betabinwidth");
  nbetabins = intParameter("nbetabins");
  betabinref = realParameter("betabinref");
  betamin = realParameter("betamin");
  betamax = realParameter("betamax");
  betaminwarn = realParameter("betaminwarn");
  betamaxwarn = realParameter("betamaxwarn");
  xdispbinning = stringParameter("xdispbinning");
  nxdispbins = intParameter("nxdispbins");
  xdispbinwidth = realParameter("xdispbinwidth");
  xdispbinref = realParameter("xdispbinref");
  xdispmin = realParameter("xdispmin");
  xdispmax = realParameter("xdispmax");
  xdispminwarn = realParameter("xdispminwarn");
  xdispmaxwarn = realParameter("xdispmaxwarn");
  withpointingcolumn = booleanParameter("withpointingcolumn");
  withmlambdacolumn = booleanParameter("withmlambdacolumn");

// gtimerge
  withauxgti = booleanParameter("withauxgti");
  gtitable = stringParameter("gtitable");

// rgsfilter
  rejflags = stringListParameter("rejflags");
  withcombmap = booleanParameter("withcombmap");

// rgsregions
  spatialres = intParameter("spatialres");
  orderres = intParameter("orderres");
  xpsfincl = realParameter("xpsfincl");
  xpsfexcl = realParameter("xpsfexcl");
  pdistincl = realParameter("pdistincl");
  procsrcsexpr = stringParameter("procsrcsexpr");
  exclsrcsexpr = stringParameter("exclsrcsexpr");

// rgsspectrum
  rebin = intParameter("rebin");
  edgechannels = intParameter("edgechannels");

  int mask = 0;
  if ( processevents ) mask |= 1;
  if ( processsources ) mask |= 2;
  if ( processexposure ) mask |= 4;
  if ( processregions ) mask |= 8;
  if ( mask==0 || mask==5 || mask==9 || mask==10 || mask==11 || mask==13 )
    errstr << "cannot leave out a middle processing stage"
           << Sas::error("middleProcessMissing");
 
  // Make a list of instruments and exposures to process.
  list<Xmm::Instrument> instruments;
  for(int i = 0; i < parameterCount("instruments"); i++)
	instruments.push_back(Xmm::toInstrument(stringParameter("instruments",i)));
  
  Xmm::DataMode diagnosticIds[] = { Xmm::DATA_DIAGNOSTIC};
  list<Xmm::DataMode> diamodes(diagnosticIds,
	diagnosticIds + sizeof(diagnosticIds)/sizeof(diagnosticIds[0]));
  
  Xmm::DataMode modeIds[] = { Xmm::DATA_SPECTROSCOPY, Xmm::HIGH_TIME_RESOLUTION };
  list<Xmm::DataMode> modes(modeIds,
	modeIds + sizeof(modeIds)/sizeof(modeIds[0]));

  const Odf *odf = oalServer.odf();
  errstr << odf->proposalInfo() << Sas::message(Msg::Meta,Msg::Verbose);

  // Construct file basename for this ODF.
  const OdfSet *summarySet = odf->selectSet(Xmm::SC,Xmm::SUMMARY_INFORMATION,1);
  string odfBasename = basename(summarySet);

  // run atthkgen if no attitude was specified
  if ( !withattitude ) {
     attitude = odfBasename + "_attitude.ds";
     if ( processevents ) {
	openParameters("atthkgen");
	setParameter("timestep",timestep);
	setParameter("atthkset",attitude);
	task(atthkgen);
	closeParameters();
     }
  }

  for (list<Xmm::Instrument>::const_iterator ii = instruments.begin();
			ii != instruments.end(); ++ii) {

    oalServer.instrument(*ii);

    string instrumentBasename = basename(summarySet,(*ii));

    // construct list of diagnostic mode datasets

    oalServer.dataMode(Xmm::DATA_DIAGNOSTIC);
    vector<string> insets;

    for (unsigned exp = 1; exp <= odf->exposures(); ++exp) {
      errstr << "Diagnostic Exposure " << exp
             << Sas::message(Msg::Meta,Msg::Verbose);
      oalServer.exposureIndex(exp);

      const OdfSet *diags = odf->selectSet(Xmm::DATA_DIAGNOSTIC);
      insets.push_back(diags->name());
    }

    string offsetset = instrumentBasename + "_offset.ds";
    if (processevents && calcoffsets) {
      // run rgsoffsetcalc
      openParameters("rgsoffsetcalc");
      setParameter("sets", insets);
      setParameter("outset", offsetset);
      setParameter("withhistogramset", withoffsethistogram);
      setParameter("histogramset", instrumentBasename + "_offsethisto.ds");
      task(rgsoffsetcalc);
      closeParameters();
    }

    for (list<Xmm::DataMode>::const_iterator im = modes.begin();
			im != modes.end(); ++im) {
      oalServer.dataMode(*im);
      errstr << "Instrument " << int(*ii) << " mode " << int(*im)
             << " exposures " << odf->exposures()
             << Sas::message(Msg::Meta,Msg::Verbose);

      for (unsigned exp = 1; exp <= odf->exposures(); ++exp) {
        errstr << "Exposure " << exp << Sas::message(Msg::Meta,Msg::Verbose);
	oalServer.instrument(*ii);
	oalServer.dataMode(*im);
        oalServer.exposureIndex(exp);
        const Exposure* exposure = odf->currentExposure();

	errstr << *exposure << Sas::message(Msg::Meta,Msg::Noisy);

	process(exposure,offsetset);
      }
    }
  }

  ENDPROC
}
